"""ResellerClub API URLs"""


from .base import BaseURLs
from .domains import DomainsURLs
from .customers import CustomersURLs


class URLs(BaseURLs):
    """Stores all API URLs"""

    prod_url = "https://httpapi.com/api/"
    test_url = "https://test.httpapi.com/api/"

    def __init__(self, test_mode: bool = True) -> None:
        """Stores all API URLs

        Args:
            debug (bool, optional): Use the test or live API URLs. Defaults to True.
            response_format (str, optional): Valid values are json or xml. Defaults to "json".
        """
        base_url = self.test_url if test_mode else self.prod_url
        super().__init__(base_url)

        # Domains urls
        self.domains = DomainsURLs(self.base_url)
        self.customers = CustomersURLs(self.base_url)



"""Base URL Classes"""


class BaseURLs:
    """URLs base class"""

    base_url = ""

    def __repr__(self) -> str:
        return f"<{self.__class__.__name__}: {self}>"

    def __str__(self) -> str:
        return self.base_url

    def __init__(self, base_url: str = "") -> None:
        """Stores URLs

        Args:
            base_url (str, optional): Base url for all the endpoints. Defaults to "".
            response_format (str, optional): Valid values are json or xml. Defaults to "json".
        """
        self.base_url = f"{base_url}{self.base_url}"



"""Customers API Client"""
from datetime import datetime
from typing import Iterator, List, Literal, NamedTuple

from .base import BaseClient


class Customer(NamedTuple):
    """Customer object"""

    id: str
    username: str
    reseller_id: str
    name: str
    company: str
    city: str
    state: str
    country: str
    status: str
    total_receipts: float
    phone: str
    phone_country_code: str
    website_count: int


class SearchResponse(NamedTuple):
    """Represents the result of a customer search"""

    page_records: int
    db_records: int
    customers: List[Customer]

    def __len__(self) -> int:
        return len(self.customers)

    def __iter__(self) -> Iterator:
        return self.customers.__iter__()


class CustomersClient(BaseClient):
    """Customers API Client"""

    def search(
        self,
        records: int,
        page: int,
        customers: List[str] | str = None,
        resellers: List[str] | str = None,
        username: str = None,
        name: str = None,
        company: str = None,
        city: str = None,
        state: str = None,
        status: Literal["Active", "Suspended", "Deleted"] = None,
        creation_date_start: datetime = None,
        creation_date_end: datetime = None,
        total_receipt_start: float = None,
        total_receipt_end: float = None,
    ) -> SearchResponse[Customer]:
        """Gets details of the Customers that match the search criteria

        Args:
            records (int): Number of records to be fetched.
            page (int): Page number for which details are to be fetched
            customers (List[str] | str, optional): Customer ID(s). Defaults to None.
            resellers (List[str] | str, optional): Reseller ID(s) for whom Customer accounts need
            to be searched. Defaults to None.
            username (str, optional): Username of Customer. Should be an email address.
            Defaults to None.
            name (str, optional): Name of Customer. Defaults to None.
            company (str, optional): Comany name of Customer. Defaults to None.
            city (str, optional): City. Defaults to None.
            state (str, optional): State. Defaults to None.
            status (Literal["Active", "Suspended", "Deleted"], optional): Status of Customer.
            Defaults to None.
            creation_date_start (datetime, optional): DateTime for listing of Customer accounts
            whose Creation Date is greater than. Defaults to None.
            creation_date_end (datetime, optional): DateTime for listing of Customer accounts whose
            Creation Date is less than. Defaults to None.
            total_receipt_start (float, optional): Total receipts of Customer which is greater than.
            Defaults to None.
            total_receipt_end (float, optional): Total receipts of Customer which is less than.
            Defaults to None.

        Returns:
            SearchResponse[Customer]: Object containing Customer objects for each client matching
            search criteria
        """

        if isinstance(customers, str):
            customers = [customers]
        if isinstance(resellers, str):
            resellers = [resellers]
        if creation_date_start:
            creation_date_start = creation_date_start.timestamp()
        if creation_date_end:
            creation_date_end = creation_date_end.timestamp()

        url = self._urls.customers.get_search_url()
        params = {
            "no-of-records": records,
            "page-no": page,
            "customer-id": customers,
            "reseller-id": resellers,
            "username": username,
            "name": name,
            "company": company,
            "city": city,
            "state": state,
            "status": status,
            "creation-date-start": creation_date_start,
            "creation-date-end": creation_date_end,
            "total-receipt-start": total_receipt_start,
            "total-receipt-end": total_receipt_end,
        }
        data = self._get_data(url, params)

        recsonpage = int(data.get("recsonpage"))
        recsindb = int(data.get("recsindb"))
        customers = []
        for key, value in data.items():
            if key.isdigit():
                customer_data = {k.split(".")[1]: v for k, v in value.items()}
                customer_params = {
                    "id": customer_data["customerid"],
                    "username": customer_data["username"],
                    "reseller_id": customer_data["resellerid"],
                    "name": customer_data["name"],
                    "company": customer_data["company"],
                    "city": customer_data["city"],
                    "state": customer_data.get("state"),
                    "country": customer_data["country"],
                    "status": customer_data["customerstatus"],
                    "total_receipts": float(customer_data["totalreceipts"]),
                    "phone": customer_data["telno"],
                    "phone_country_code": customer_data["telnocc"],
                    "website_count": int(customer_data["websitecount"]),
                }
                customers.append(Customer(**customer_params))

        return SearchResponse(recsonpage, recsindb, customers)



"""ResellerClub API Client"""

from .client.customers import CustomersClient
from .client.domains import DomainsClient


class ResellerClub:
    """ResellerClub API Client"""

    def __init__(
        self,
        auth_userid: str,
        api_key: str,
        test_mode: bool = True,
    ) -> None:

        self.domains = DomainsClient(auth_userid, api_key, test_mode)
        self.customers = CustomersClient(auth_userid, api_key, test_mode)



"""Base API classes"""
import requests

from ..exceptions import ResellerClubAPIException
from .urls import URLs


class BaseClient:
    """Base API Client class"""

    def __init__(
        self,
        auth_userid: str,
        api_key: str,
        test_mode: bool = True,
    ) -> None:
        self._auth_userid = auth_userid
        self._api_key = api_key
        self._urls = URLs(test_mode)

    def _get_data(self, url: str, params: dict = None) -> dict:
        """Get response from API

        Args:
            url (str): URL to request data from
            params (dict, optional): Parameters of the request. Defaults to None.

        Returns:
            dict: dict with response data
        """
        params.update({"auth-userid": self._auth_userid, "api-key": self._api_key})
        response = requests.get(url, params, timeout=120)

        try:
            data = response.json()
        except requests.JSONDecodeError:
            response.raise_for_status()

        if not response.ok:
            raise ResellerClubAPIException(data["message"])

        return data



"""Domains API Client"""
from typing import List, NamedTuple

from .base import BaseClient


class Availability(NamedTuple):
    """Domain name availability for TLDs"""

    domain: str
    status: str
    classkey: str = None


class PremiumDomain(NamedTuple):
    """Premium Domain"""

    domain: str
    price: float


class Suggestion(NamedTuple):
    """Domain name suggestion"""

    domain: str
    status: str
    in_ga: bool
    score: float
    spin: str


class DomainsClient(BaseClient):
    """Domains API Client. Methods to Search, Register or Renew domain names, etc."""

    def check_availability(self, domain_names: list, tlds: list) -> List[Availability]:
        """Checks the availability of the specified domain name(s).
        https://manage.resellerclub.com/kb/answer/764

        Args:
            domain_names (list): Domain name(s) that you need to check the availability for
            tlds (list): TLDs for which the domain name availability needs to be checkedW

        Returns:
            List[Availability]: Returns a list containing domain name availability status for the
            requested TLDs
        """
        params = {"domain-name": domain_names, "tlds": tlds}
        url = self._urls.domains.get_availability_check_url()
        data = self._get_data(url, params)

        return [Availability(dn, **a) for dn, a in data.items() if not dn == "errors"]

    def check_idn_availability(
        self, domain_names: list, tld: str, idn_language_code: str
    ) -> List[Availability]:
        """Checks the availability of the specified Internationalized Domain Name(s) (IDN)
        https://manage.resellerclub.com/kb/answer/1427

        Args:
            domain_names (list): Internationalized Domaine Name(s) that you need to check the
            availability for
            tld (str): TLD for which the domain name availability needs to be checked
            idn_language_code (str): While performing check availability for an Internationalized
            Domain Name, you need to provide the corresponding language code

        Returns:
            List[Availability]: List containing domain name availability status for the requested
            TLDs
        """
        params = {
            "domain-name": domain_names,
            "tld": tld,
            "idnLanguageCode": idn_language_code,
        }
        url = self._urls.domains.get_availability_check_url("idn")
        data = self._get_data(url, params)

        return [Availability(dn, **availability) for dn, availability in data.items()]

    def check_premium_domain_availability(
        self,
        keyword: str,
        tlds: list,
        highest_price: int = None,
        lowest_price: int = None,
        max_results: int = None,
    ) -> List[PremiumDomain]:
        """Returns a list of Aftermarket Premium domain names based on the specified keyword.
        This method only returns names available on the secondary market, and not those premium
        names that are offered directly by any Registry for new registration.
        https://manage.resellerclub.com/kb/answer/1948

        Args:
            keyword (str): Word or phrase (please enter the phrase without spaces) for which
            premium search is requested
            tlds (list): Domain name extensions (TLDs) you want to search in
            highest_price (int, optional): Maximum price (in Reseller's Selling Currency) up to
            which domain names must be suggested. Defaults to None.
            lowest_price (int, optional): Minimum price (in Reseller's Selling Currency) for which
            domain names must be suggested. Defaults to None.
            max_results (int, optional): Number of results to be returned. Defaults to None.

        Returns:
            List[PremiumDomain]: List of domain names and prices
        """

        params = {
            "key-word": keyword,
            "tlds": tlds,
            "price-high": highest_price,
            "price-low": lowest_price,
            "no-of-results": max_results,
        }
        url = self._urls.domains.get_availability_check_url("premium")
        data = self._get_data(url, params)

        return [PremiumDomain(domain, float(price)) for domain, price in data.items()]

    def check_third_level_name_availability(
        self, domain_names: list
    ) -> List[Availability]:
        """Checks the availability of the specified 3rd level .NAME domain name(s).
        https://manage.resellerclub.com/kb/node/2931

        Args:
            domain_names (list): Domain name(s) that you need to check the availability for.

        Returns:
            List[Availability]: List containing domain name availability status for the requested
            domain names
        """
        params = {"domain-name": domain_names, "tlds": "*.name"}
        url = self._urls.domains.get_availability_check_url("3rd_level_dotname")
        data = self._get_data(url, params)

        return [Availability(dn, **availability) for dn, availability in data.items()]

    def suggest_names(
        self,
        keyword: str,
        tld_only: str = None,
        exact_match: bool = None,
        adult: bool = None,
    ) -> List[Suggestion]:
        """Returns domain name suggestions for a user-specified keyword.
        https://manage.resellerclub.com/kb/answer/1085

        Args:
            keyword (str): Search term (keyword or phrase) e.g. "search" or "search world"
            tld_only (str, optional): Specific TLD(s) you may want to search for. Defaults to None.
            exact_match (bool, optional): Will return keyword alternatives when set to True.
            Can be set to False to only return TLD alternatives. Defaults to None.
            adult (bool, optional): If set to false, the suggestions will not contain any adult or
            explicit suggestions which contain words like "nude", "porn", etc. Defaults to None.

        Returns:
            List[Suggestion]: List of domain name suggestions.
        """
        params = {
            "keyword": keyword,
            "tld-only": tld_only,
            "exact-match": exact_match,
            "adult": adult,
        }
        url = self._urls.domains.get_name_suggestion_url()
        data = self._get_data(url, params)

        result = []
        for domain, sug in data.items():
            in_ga = bool(sug["in_ga"].lower() == "true")
            score = float(sug["score"])
            suggestion_params = [domain, sug["status"], in_ga, score, sug["spin"]]
            result.append(Suggestion(*suggestion_params))

        return result



"""Domains Unit Tests"""
import idna
from case import ResellerClubTestCase
from thefuzz import fuzz


class TestDomainAvailability(ResellerClubTestCase):
    """Domain Availability Test Cases"""

    domains = ["github", "google"]
    tlds = ["com", "net"]

    def test_single_domain_single_tld(self):
        """Test single domain with single TLD case"""
        domain = self.domains[0]
        tld = self.tlds[0]
        result = self.api.domains.check_availability([domain], [tld])

        expected_domain = f"{domain}.{tld}"
        self.assertIn(expected_domain, [availability.domain for availability in result])

    def test_single_domain_multiple_tlds(self):
        """Test single domain with multiple TLDs case"""
        domain = self.domains[0]
        tlds = self.tlds
        result = self.api.domains.check_availability([domain], tlds)

        expected_domains = [f"{domain}.{tld}" for tld in tlds]
        result_domains = [a.domain for a in result]
        self.assertListEqual(sorted(expected_domains), sorted(result_domains))

    def test_multiple_domains_single_tld(self):
        """Test multiple domains with single TLD case"""
        domains = self.domains
        tld = self.tlds[0]
        result = self.api.domains.check_availability(domains, [tld])

        expected_domains = [f"{domain}.{tld}" for domain in domains]
        result_domains = [a.domain for a in result]
        self.assertListEqual(sorted(expected_domains), sorted(result_domains))

    def test_multiple_domains_multiple_tlds(self):
        """Test multiple domains with multiple TLDs case"""
        domains = self.domains
        tlds = self.tlds
        result = self.api.domains.check_availability(domains, tlds)

        expected_domains = [f"{domain}.{tld}" for domain in domains for tld in tlds]
        result_domains = [a.domain for a in result]
        self.assertListEqual(sorted(expected_domains), sorted(result_domains))


class TestIDNAvailability(ResellerClubTestCase):
    """IDN Availability Test Cases"""

    domains = ["ѯҋ111", "ѯҋ112"]
    tld = "com"
    idn_language_code = "aze"

    def test_single_domain(self):
        """Test single IDN case"""
        domain = self.domains[0]
        tld = self.tld
        result = self.api.domains.check_idn_availability(
            [domain], tld, self.idn_language_code
        )

        punycode_domain = idna.encode(domain).decode()
        expected_domain = f"{punycode_domain}.{tld}"
        self.assertIn(expected_domain, [a.domain for a in result])

    def test_multiple_domains(self):
        """Test multiple IDNs case"""
        domains = self.domains
        tld = self.tld
        result = self.api.domains.check_idn_availability(
            domains, tld, self.idn_language_code
        )

        punycode_domains = [idna.encode(domain).decode() for domain in domains]
        expected_domains = [f"{domain}.{tld}" for domain in punycode_domains]
        result_domains = [a.domain for a in result]
        self.assertListEqual(sorted(expected_domains), sorted(result_domains))


class TestPremiumDomainsAvailability(ResellerClubTestCase):
    """Premium domains availability check test case"""

    keyword = "domain"
    tlds = ["com", "net", "org"]
    highest_price = 10000
    lowest_price = 100
    max_results = 10

    def test_single_tld(self):
        """Test single TLD case"""
        keyword = self.keyword
        tld = self.tlds[0]

        response = self.api.domains.check_premium_domain_availability(keyword, [tld])

        domains = [pd.domain for pd in response]
        is_keyword_in_domains = all(keyword in domain for domain in domains)
        is_tld_in_domains = all(domain.endswith(f".{tld}") for domain in domains)

        self.assertTrue(is_keyword_in_domains, "Keyword is not in response")
        self.assertTrue(is_tld_in_domains, "TLD is not in response")

    def test_multiple_tlds(self):
        """Test multiple TLDs case"""
        keyword = self.keyword
        tlds = self.tlds

        response = self.api.domains.check_premium_domain_availability(keyword, tlds)

        domains = [pd.domain for pd in response]
        is_keyword_in_domains = all(keyword in key for key in domains)
        is_tld_in_domains = all(any(d.endswith(f".{t}") for d in domains) for t in tlds)

        self.assertTrue(is_keyword_in_domains, "Keyword is not in response")
        self.assertTrue(is_tld_in_domains, "TLD is not in response")

    def test_highest_price(self):
        """Test highest price case"""
        highest_price = self.highest_price
        params = [self.keyword, self.tlds[0], highest_price]
        response = self.api.domains.check_premium_domain_availability(*params)
        prices = [pd.price for pd in response]

        self.assertGreaterEqual(highest_price, max(prices))

    def test_lowest_price(self):
        """Test lowest price case"""
        lowest_price = self.lowest_price
        params = [self.keyword, self.tlds[0], None, lowest_price]
        response = self.api.domains.check_premium_domain_availability(*params)
        prices = [pd.price for pd in response]

        self.assertGreaterEqual(min(prices), lowest_price)

    def test_highest_and_lowest_price(self):
        """Test highest and lowest price case"""
        lowest_price = self.lowest_price
        highest_price = self.highest_price
        params = [self.keyword, self.tlds[0], highest_price, lowest_price]
        response = self.api.domains.check_premium_domain_availability(*params)
        prices = [pd.price for pd in response]

        self.assertGreaterEqual(min(prices), lowest_price)
        self.assertGreaterEqual(highest_price, max(prices))

    def test_max_results(self):
        """Test max results case"""
        max_results = self.max_results
        params = {
            "keyword": self.keyword,
            "tlds": self.tlds[0],
            "max_results": max_results,
        }
        response = self.api.domains.check_premium_domain_availability(**params)

        self.assertGreaterEqual(max_results, len(response))


class TestThirdLvlNameAvailability(ResellerClubTestCase):
    """.NAME 3rd level availability check test case"""

    domains = ["domain.one", "domain.two"]

    def test_single_domain(self):
        """Test single domain case"""
        result = self.api.domains.check_third_level_name_availability(self.domains[0])

        expected_domain = f"{self.domains[0]}.name"
        result_domains = [a.domain for a in result]
        self.assertListEqual([expected_domain], sorted(result_domains))

    def test_multiple_domain(self):
        """Test multiple domain case"""
        result = self.api.domains.check_third_level_name_availability(self.domains)

        expected_domains = [f"{domain}.name" for domain in self.domains]
        result_domains = [a.domain for a in result]
        self.assertListEqual(sorted(expected_domains), sorted(result_domains))


class TestSuggestNames(ResellerClubTestCase):
    """Suggest name test case"""

    keyword = "reseller"

    def test_keyword_only(self):
        """Test suggest names with keyword only"""
        suggestions = self.api.domains.suggest_names(self.keyword)
        ratio_list = [fuzz.partial_ratio(s.domain, self.keyword) for s in suggestions]
        failed_message = "Some results are less than 75% similar"
        self.assertFalse(any(ratio < 75 for ratio in ratio_list), failed_message)

    def test_tld(self):
        """Test suggest names with keyword and .com tld"""
        tld = "com"
        suggestions = self.api.domains.suggest_names(self.keyword, tld)
        assertion = all(s.domain.endswith(f".{tld}") for s in suggestions)
        self.assertTrue(assertion, "Some results do not contain TLD")

    def test_exact_match(self):
        """Test suggest names with keyword exact match"""
        suggestions = self.api.domains.suggest_names(self.keyword, exact_match=True)
        assertion = all(s.domain.split(".")[0] == self.keyword for s in suggestions)
        self.assertTrue(assertion, "Results are not exact match")



"""Reseller Club API Test Case"""
from unittest import TestCase

from settings import API_KEY, RESELLER_ID

from src.resellerclub import ResellerClub


class ResellerClubTestCase(TestCase):
    """Base ResellerClub API Test Case"""

    api = ResellerClub(RESELLER_ID, API_KEY)



"""Domains URL endpoints"""
from typing import Literal

from .base import BaseURLs


class DomainsURLs(BaseURLs):
    """Stores all API URLs to search, register or renew domain names"""

    base_url = "domains/"

    def get_availability_check_url(
        self, domain_type: Literal["idn", "premium", "3rd_level_dotname"] = None
    ) -> str:
        """Returns URL for domain availability check

        Args:
            domain_type: Type of domain to check.
                "idn": Internationalized Domain Name availability check.
                "premium": Premium Domain Availability check.
                "3rd_level_dotname": 3rd level .NAME availability check.
                None: Regular domain availability check.
                Defaults to None.

        Returns:
            str: API endpoint URL
        """

        url_map = {
            "idn": f"{self.base_url}idn-available.json",
            "premium": f"{self.base_url}premium/available.json",
            "3rd_level_dotname": f"{self.base_url}thirdlevelname/available.json",
        }

        return url_map.get(domain_type, f"{self.base_url}available.json")

    def get_name_suggestion_url(self) -> str:
        """Returns URL for domain name suggestion

        Returns:
            str: API endpoint URL
        """
        return f"{self.base_url}v5/suggest-names.json"



"""Customers URL endpoints"""
from .base import BaseURLs


class CustomersURLs(BaseURLs):
    """Sets up the URLs for the customers endpoints"""

    base_url = "customers/"

    def get_search_url(self) -> str:
        """Returns the URL for the search endpoint"""
        return f"{self.base_url}search.json"



"""Customers Unit Tests"""
from case import ResellerClubTestCase


class TestSearchCustomers(ResellerClubTestCase):
    """Test SearchCustomers"""

    def __init__(self, methodName: str = "runTest") -> None:
        super().__init__(methodName)
        self.search_method = self.api.customers.search

    def test_search_first_ten_customers(self):
        """Test search first ten customers"""
        customers = self.search_method(10, 1)
        self.assertGreaterEqual(10, len(customers))
